/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: dcf77.h,v 1.24 2005/09/09 21:02:58 phk Exp $
 */


#include <sys/queue.h>

#define DCF77_MAGIC ~77500U

extern struct logh *l_msg;

enum state {
	DCF_UNCONFIG = 0,
	DCF_INIT,
	DCF_SEARCH,
	DCF_MINUTE,
	DCF_HOUR,
	DCF_DAY,
	DCF_SYNC,
	DCF_LOCK
};

struct dcf77 {
	u_int		magic;
	struct ck_instance cip;
	u_int		unit;

	struct ppsdev	*pps;

	struct source	*sid;

	/* Tweakables */
	double		max_gap;
	double		max_jitter;
	double		low_zero;
	double		high_zero;
	double		low_one;
	double		high_one;
	double		disproof;
	double		timeout;
	double		squelch;
	double		ts_zero;
	double		ts_one;
	double		lockbits;
	double		untrack;
	double		delay;
#define NPREAMBLE	15
	int		preamble[NPREAMBLE];

	struct timespec	sample_zero;
	struct timespec	sample_one;

	enum state	state;

	evTimerID	tid;
	enum {
	    SHORT,
	    LONG
	}		timer_state;

	struct ts	ts[3];
	int		missing;
	int		minute;
	int		field[17];
	int		parity[4];
	time_t		guess_t;
	struct tm	guess_tm;


#define NDWHIST		71
#define NDWBUCKET	5e-3
	u_int		dwhist[NDWHIST];
#define NDTHIST		61
	u_int		dthist[NDTHIST];

	int		shiftreg[60];
	int		shiftprev[60];
	int		shiftlast[60];
	int		lasti;
	int		target;
	uint64_t	candidate;	/*
					 * Candidates are the positions which
					 * currently have no votes against
					 * them
					 */
	uint64_t	possible;	/*
					 * Possible are the positions which
					 * have not had more than the threshold
					 * votes against them at any time
					 */
	time_t		t0;
	int		noise;

	int		synth[60];	/*
					 * Synthesized timecode in tracking
					 * mode.
					 */
	int		corr[60];	/*
					 * match/fail status of each bit
					 */
	int		match;
	int		fail;
	time_t		delta;

	int		clock_step;

};

/* blame.c */
int blameallocation(struct dcf77 *ip, int offset, int limit, char *why);

/* timegram.c */
void dcf77_change_state(struct dcf77 *ip, enum state newstate);
void dcf77_addpulse(struct dcf77 *ip, int sec, int class);
