/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 *
 * $Id: dcf77_timegram.c,v 1.16 2005/10/04 07:25:43 phk Exp $
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <configkit.h>
#include <isc/eventlib.h>
#include <ntpns.h>
#include "dcf77.h"

static const char *
state_name(enum state st)
{
	switch(st) {
	case DCF_UNCONFIG:	return("unconfig");
	case DCF_INIT:		return("init");
	case DCF_SEARCH:		return("search");
	case DCF_MINUTE:		return("minute");
	case DCF_HOUR:		return("hour");
	case DCF_DAY:		return("day");
	case DCF_SYNC:		return("sync");
	case DCF_LOCK:		return("lock");
	default:		return("st=?");
	}
}

/**********************************************************************/
static void
dcf77_synth(struct dcf77 *ip, time_t t)
{
	struct tm mtm;
	int i, j, k;

	t += 60;
	t -= t % 60;

	memset(ip->synth, 0, sizeof ip->synth);
	localtime_r(&t, &mtm);

	/* Use our configured preamble */
	for (i = 0; i < NPREAMBLE; i++)
		ip->synth[i] = ip->preamble[i];

	/* R: no idea */
	ip->synth[i++] = 0;

	/* A1: Only possible in early hours of sunday */
	if (mtm.tm_hour > 3 || mtm.tm_wday != 0)
		ip->synth[i++] = -1;
	else
		ip->synth[i++] = 0;

	/* Z1 + Z2 */
	ip->synth[i++] = mtm.tm_isdst ? 1 : -1;
	ip->synth[i++] = mtm.tm_isdst ? -1 : 1;

	/* A2: Only possible in early hours on the 1st of the month */
	if (mtm.tm_hour > 3 || mtm.tm_mday != 1)
		ip->synth[i++] = -1;
	else
		ip->synth[i++] = 0;

	/* S: always one */
	ip->synth[i++] = 1;

	/* Minute ones */
	j = mtm.tm_min % 10;
	ip->synth[i++] = j & 1 ? 1 : -1;
	ip->synth[i++] = j & 2 ? 1 : -1;
	ip->synth[i++] = j & 4 ? 1 : -1;
	ip->synth[i++] = j & 8 ? 1 : -1;

	/* Minute tens */
	j = mtm.tm_min / 10;
	ip->synth[i++] = j & 1 ? 1 : -1;
	ip->synth[i++] = j & 2 ? 1 : -1;
	ip->synth[i++] = j & 4 ? 1 : -1;

	/* P1 */
	k = -1;
	for (j = 21; j < 28; j++)
		k *= -ip->synth[j];
	ip->synth[i++] = k;

	/* Hour ones */
	j = mtm.tm_hour % 10;
	ip->synth[i++] = j & 1 ? 1 : -1;
	ip->synth[i++] = j & 2 ? 1 : -1;
	ip->synth[i++] = j & 4 ? 1 : -1;
	ip->synth[i++] = j & 8 ? 1 : -1;

	/* Hour tens */
	j = mtm.tm_hour / 10;
	ip->synth[i++] = j & 1 ? 1 : -1;
	ip->synth[i++] = j & 2 ? 1 : -1;

	/* P2 */
	k = -1;
	for (j = 29; j < 35; j++)
		k *= -ip->synth[j];
	ip->synth[i++] = k;

	/* Day ones */
	j = mtm.tm_mday % 10;
	ip->synth[i++] = j & 1 ? 1 : -1;
	ip->synth[i++] = j & 2 ? 1 : -1;
	ip->synth[i++] = j & 4 ? 1 : -1;
	ip->synth[i++] = j & 8 ? 1 : -1;

	/* Day tens */
	j = mtm.tm_mday / 10;
	ip->synth[i++] = j & 1 ? 1 : -1;
	ip->synth[i++] = j & 2 ? 1 : -1;

	/* Week day */
	j = mtm.tm_wday;
	if (j == 0)
		j = 7;
	ip->synth[i++] = j & 1 ? 1 : -1;
	ip->synth[i++] = j & 2 ? 1 : -1;
	ip->synth[i++] = j & 4 ? 1 : -1;

	/* Month ones */
	j = (mtm.tm_mon + 1) % 10;
	ip->synth[i++] = j & 1 ? 1 : -1;
	ip->synth[i++] = j & 2 ? 1 : -1;
	ip->synth[i++] = j & 4 ? 1 : -1;
	ip->synth[i++] = j & 8 ? 1 : -1;

	/* Month tens */
	j = (mtm.tm_mon + 1) / 10;
	ip->synth[i++] = j & 1 ? 1 : -1;

	/* Year ones */
	j = mtm.tm_year % 10;
	ip->synth[i++] = j & 1 ? 1 : -1;
	ip->synth[i++] = j & 2 ? 1 : -1;
	ip->synth[i++] = j & 4 ? 1 : -1;
	ip->synth[i++] = j & 8 ? 1 : -1;

	/* Year tens */
	j = (mtm.tm_year / 10) % 10;
	ip->synth[i++] = j & 1 ? 1 : -1;
	ip->synth[i++] = j & 2 ? 1 : -1;
	ip->synth[i++] = j & 4 ? 1 : -1;
	ip->synth[i++] = j & 8 ? 1 : -1;

	/* P3 */
	k = -1;
	for (j = 36; j < 58; j++)
		k *= -ip->synth[j];
	ip->synth[i++] = k;
}

/**********************************************************************/

void
dcf77_change_state(struct dcf77 *ip, enum state newstate)
{

	Log(l_msg, "State %s -> %s",
	    state_name(ip->state), state_name(newstate));
	ip->state = newstate;
	switch (ip->state) {
	case DCF_INIT:
		/* Discard PPS data to avoid an initial very long pulse */
		(void)PpsPickup(&ip->pps, ip->ts, 2);
		/* Use SHORT mode until we find a flank */
		ip->timer_state = SHORT;
		/* Clear everything */
		memset(ip->shiftreg, 0, sizeof ip->shiftreg);
		memset(ip->shiftprev, 0, sizeof ip->shiftprev);
		memset(ip->shiftlast, 0, sizeof ip->shiftlast);
		memset(ip->corr, 0, sizeof ip->corr);
		ip->candidate = (1LL << 60LL) - 1;
		ip->possible = (1LL << 60LL) - 1;
		ip->noise = 0;
		ip->missing = 0;
		ip->lasti = -1;
		ip->guess_t = 0;
		ip->match = 0;
		ip->fail = 0;
		ip->delta = 0;
		ip->minute = 0;
		time(&ip->t0);
		dcf77_synth(ip, ip->t0);
		dcf77_change_state(ip, DCF_SEARCH);
		break;
	case DCF_SEARCH:
	case DCF_MINUTE:
	case DCF_HOUR:
	case DCF_DAY:
	case DCF_SYNC:
	case DCF_LOCK:
		break;
	default:
		Log(l_msg, "Not handled, state %s", state_name(ip->state));
		dcf77_change_state(ip, DCF_INIT);
		break;
	}
}


static int
dcf77_pulsechar(int j)
{

	if (j < 0)
		return ('-');
	if (j > 0)
		return ('+');
	return ('_');
}

static const char *
dcf77_pulsetrain(int *r, int off)
{
	static char str[61];
	int i;

	for (i = 0; i < 60; i++)
		str[i] = dcf77_pulsechar(r[(i + off) % 60]);
	str[60] = '\0';
	return (str);
}

/**********************************************************************/
static void
dcf77_search(struct dcf77 *ip, int sec, int class __unused)
{
	int i, j, reject;
	char why[BUFSIZ], lbuf[BUFSIZ];
	uint64_t bitpos;
	time_t t;

	strcpy(lbuf, "   ");
	for (i = 0; i < 60; i += 1) {
		bitpos = 1LL << (int64_t)i;

		/* Only check the current and all known possible bits */
		if (i != sec && !(ip->possible & bitpos))
			continue;

		reject = blameallocation(ip, i, ip->disproof, why);
		if (reject == 0) {
			ip->candidate |= bitpos;
			ip->possible |= bitpos;
			ip->target = i;
		} else if (reject >= ip->disproof) {
	 		if (ip->possible & bitpos)
				sprintf(lbuf + strlen(lbuf), "%d:%s, ", i, why);
			ip->possible &= ~bitpos;
			ip->candidate &= ~bitpos;
		} else {
			ip->possible |= bitpos;
		}
	}
	if (class && LogActive(l_msg)) {
		strcpy(why, "   ");
		for(i = 0; i < 60; i++) {
			if (ip->candidate & (1LL << (int64_t)i))
				strcat(why, "C");
			else if (ip->possible & (1LL << (int64_t)i))
				strcat(why, "p");
			else
				strcat(why, "_");
		}
		Log(l_msg, "%s", why);
		Log(l_msg, "%s", lbuf);
	}
	if (ip->possible == 0) { 
		Log(l_msg, "%d No cigar, try again", ip->t0);
		dcf77_change_state(ip, DCF_INIT);
		return;
	}

	j = 0;
	for (i = 0; i < 59; i++)
		if (ip->shiftlast[(ip->target + i) % 60] == 0)
			j++;
	time(&t);
	t -= ip->t0;
	
	if (j > ip->timeout * ip->squelch) {
		Log(l_msg, "%d Too noisy/silent, try again missing %d",
		    ip->t0, ip->missing);
		dcf77_change_state(ip, DCF_INIT);
		return;
	}

	if (ip->possible != ip->candidate) {
		/* Not converged yet / after all */
		return;
	}

	if ((ip->possible & (ip->possible - 1))) {
		/* More than one possible */
		return;
	}

	if (j > t * ip->squelch) {
		/* S/N too bad still */
		return;
	}

	Log(l_msg,
"Found %015jx target %2d after %d noise %d missing %d offset %d blank %d\n",
	    ip->possible, ip->target, t, ip->noise,
	    ip->missing, sec, j);
	time(&ip->guess_t);
	i = ip->guess_t % 60;
	ip->guess_t -= i;
	ip->guess_t -= ip->target;
	ip->guess_t += sec;
	ip->minute = ip->target;
	dcf77_change_state(ip, DCF_MINUTE);
	return;
}

/**********************************************************************/
static void
dcf77_timegram(struct dcf77 *ip, int sec)
{
	int reject, j;
	char why[BUFSIZ], lbuf[BUFSIZ];
	time_t t;

	j = (60 + sec - ip->minute) % 60;
	reject = blameallocation(ip, ip->target, ip->disproof, why);
	Log(l_msg, "%s sec %d rej: %d why=<%s> P1=%d P2=%d P3=%d TZ=%s",
	    state_name(ip->state), sec, reject, why,
	    ip->parity[1], ip->parity[2], ip->parity[3], getenv("TZ"));
	ip->guess_t++;
	localtime_r(&ip->guess_t, &ip->guess_tm);
	ip->guess_tm.tm_isdst = -1;
	if (ip->state == DCF_MINUTE && ip->parity[1] == 1 && j >= 28) {
		dcf77_change_state(ip, DCF_HOUR);
		ip->guess_tm.tm_min =
		    ip->field[6] + ip->field[7] * 10;
	}
	if (ip->state == DCF_HOUR && ip->parity[2] == 1) {
		if (ip->guess_tm.tm_min != 0 || j >= 35) {
			dcf77_change_state(ip, DCF_DAY);
			ip->guess_tm.tm_hour =
			    ip->field[8] + ip->field[9] * 10;
		}
	}
	if (ip->state == DCF_DAY && ip->parity[3] == 1) {
		if (ip->guess_tm.tm_min != 0 ||
		    ip->guess_tm.tm_hour != 0 || j == 58) {
			dcf77_change_state(ip, DCF_SYNC);
			ip->guess_tm.tm_mday =
			    ip->field[11] * 10 + ip->field[10];
			ip->guess_tm.tm_mon =
			    ip->field[14] * 10 + ip->field[13] - 1;
			ip->guess_tm.tm_year =
			    ip->field[16] * 10 + ip->field[15] + 100;
		}
	}
	ip->guess_t = mktime(&ip->guess_tm);
	localtime_r(&ip->guess_t, &ip->guess_tm);
	strftime(why, sizeof why, "%Y-%m-%d %H:%M:%S", &ip->guess_tm);
	time(&t);
	strftime(lbuf, sizeof lbuf, "%Y-%m-%d %H:%M:%S", localtime(&t));
	Log(l_msg, "%s (%s) sec = %d (%d) t = %d, guess_t = %d, delta= %d",
	    why, lbuf, sec, t % 60, t, ip->guess_t - 60, 60 + t - ip->guess_t);
	if (ip->state == DCF_SYNC) {
		ip->delta = ip->guess_t - (60 + t);
		dcf77_synth(ip, ip->delta + t);
		memset(ip->corr, 0, sizeof ip->corr);
		ip->match = 0;
		ip->fail = 0;
	}
}
/**********************************************************************/
static void
dcf77_track(struct dcf77 *ip, int sec, int class)
{
	int j;
	time_t t;

	j = (60 + sec - ip->minute) % 60;

	time(&t);

	if (!j) 
		dcf77_synth(ip, t + ip->delta);
	if (ip->corr[sec] < 0)
		ip->fail--;
	if (ip->corr[sec] > 0)
		ip->match--;
	ip->corr[sec] = ip->synth[j] * class;
	if (ip->corr[sec] < 0)
		ip->fail++;
	if (ip->corr[sec] > 0)
		ip->match++;

	Log(l_msg, "   %s +%d/-%d delta %d",
	    dcf77_pulsetrain(ip->synth, 0),
	    ip->match, ip->fail, ip->delta);
	if (ip->fail == 0 && ip->match >= ip->lockbits) {
		time(&ip->t0);
		if (ip->state != DCF_LOCK)
			dcf77_change_state(ip, DCF_LOCK);
	}
	if (ip->fail >= ip->untrack) {
		Log(l_msg, "Lost track");
		dcf77_change_state(ip, DCF_INIT);
	}
}

/**********************************************************************/
void
dcf77_addpulse(struct dcf77 *ip, int sec, int class)
{

	ip->shiftprev[sec] = ip->shiftreg[sec];
	ip->shiftreg[sec] = class;
	if (class != 0)
		ip->shiftlast[sec] = class;
	ip->lasti = sec;

	if (class && LogActive(l_msg)) {
		if (ip->state > DCF_SEARCH)
			Log(l_msg,
			    "   pppppppppppppppRAZZASmmmmMMM1hhhhHH2ddddDDwwwmmmmMyyyyYYYY3");
#if 0
		Log(l_msg, "%02d %s", sec,
		    dcf77_pulsetrain(ip->shiftlast, ip->minute));
		Log(l_msg, "%02d %s %s", ip->state,
		    dcf77_pulsetrain(ip->shiftprev, ip->minute),
		    state_name(ip->state));
#endif
		Log(l_msg, "%02d %s",
		    (60 + sec - ip->minute) % 60,
		    dcf77_pulsetrain(ip->shiftreg, ip->minute));
	}
	switch (ip->state) {
	case DCF_SEARCH:
		dcf77_search(ip, sec, class);
		break;
	case DCF_MINUTE:
	case DCF_HOUR:
	case DCF_DAY:
		dcf77_timegram(ip, sec);
		break;
	case DCF_SYNC:
	case DCF_LOCK:
		dcf77_track(ip, sec, class);
		break;
	default:
		break;
	}
}
