/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: config.c,v 1.43 2005/10/04 06:44:22 phk Exp $
 *
 * Parse the configuration and get things going.
 *
 */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <err.h>
#include <errno.h>
#include <isc/eventlib.h>
#include <sys/types.h>
#include <sys/queue.h>
#include <configkit.h>
#include <configkit_src.h>

#include <ntpns.h>
#include "ntpns_cfg.h"
#include "internal.h"

/*
 * This structure contains the status for each
 * of the sources of config info currently in use.
 */
struct cfgsrc {
	evFileID		fid;
	FILE			*fout;
	int			closing;
	int			fd;
	struct ck_src		*cfs;
	int			flags;
};

static void
EvConfig(evContext ctx __unused, void *uap, int fd, int eventmask __unused)
{
	int i;
	struct cfgsrc *cp;
	char *buf[BUFSIZ];

	cp = uap;

	i = read(fd, buf, BUFSIZ);
	if (i > 0) {
		ck_SrcFeed(cp->cfs, buf, i);
		if (!cp->closing)
			return;
	}
	i = evDeselectFD(ctx, cp->fid);
	if (i)
		err(1, "evDeselectFD(stdin)");
	Log(Ldebug, "Closing fd %d", fd);
	log_close(cp);
	if (fileno(cp->fout) == cp->fd)
		fclose(cp->fout);
	else
		close(fd);
	free(cp);
}

void
ConfigLog(struct cfgsrc *cp, const char *str)
{

	if (cp == NULL)
		return;
	fprintf(cp->fout, "%s", str);
	fflush(cp->fout);
}

static int
ConfigPrintf(CK_ARG, const char *fmt, va_list ap)
{
	struct cfgsrc *cp;

	cp = cfa->src;
	if (fmt != NULL)
		return (vfprintf(cp->fout, fmt, ap));
	if (cp->flags & CONFIG_PROMPT)
		fprintf(cp->fout, "NTPns > ");
	fflush(cp->fout);
	return (0);
} 

void
ConfigSource(FILE *fo, FILE *fi, int fd, int flags)
{
	struct cfgsrc *cp;
	int i;

	if (fd < 0) {
		if (fi == NULL)
			err(1, "must specify something to ConfigSource");
		fd = fileno(fi);
	}
	if (fo == NULL)
		fo = fdopen(fd, "w");
	if (fd < 0 || fo == NULL)
		err(1, "ConfigSource() giving up");
	
	cp = calloc(sizeof *cp, 1);
	if (cp == NULL) {
		nomem();
		fclose(fo);
		return;
	}
	cp->fd = fd;
	cp->fout = fo;
	cp->flags = flags;
	cp->cfs = ck_SrcCreate(cft, ConfigPrintf, cp);
	if (cp->flags & CONFIG_PROMPT)
		fprintf(cp->fout, "NTPns > ");
	fflush(cp->fout);
	i = evSelectFD(C, fd, EV_READ | EV_EXCEPT, EvConfig, cp, &cp->fid);
	Assert(i == 0, "evSelectFD failed in %s", __func__);
}

int
cfg_config_exit(CK_ARG)
{
	struct cfgsrc *cp;

	if (cfa->flag == 0) {
		cp = cfa->src;
		cp->closing = 1;
	}

	return (0);
}
