/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: load.c,v 1.18 2005/09/09 20:41:55 phk Exp $
 *
 * Load and unload modules
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <err.h>
#include <fcntl.h>
#include <errno.h>
#include <dlfcn.h>
#include <configkit.h>
#include <isc/eventlib.h>
#include <sys/queue.h>

#include "ntpns.h"
#include "ntpns_cfg.h"

static TAILQ_HEAD(, load) load_head =
	TAILQ_HEAD_INITIALIZER(load_head);

/*
 * This structure contains the status for each
 * loaded module
 */
struct load {
	TAILQ_ENTRY(load)	next;
	char			*name;
	void			*dlhandle;
};

int
cfg_load(CK_ARG, const char **module)
{
	mod_init_t *func;
	struct load *ld;
	const char *p;
	char *q, *r;
	int ret;

	if (cfa->flag & (CK_SHOW|CK_CONF)) {
		TAILQ_FOREACH(ld, &load_head, next) {
			ck_printf(CK_PASS, "load %s\n", ld->name);
		}
		return (0);
	}
	if (cfa->flag & CK_NO)
		return (EINVAL);

	ld = calloc(sizeof *ld, 1);
	if (ld == NULL)
		return (ENOMEM);
	ld->name = strdup(*module);
	if (ld->name == NULL) {
		free(ld);
		return (ENOMEM);
	}
	TAILQ_INSERT_TAIL(&load_head, ld, next);
	ld->dlhandle = dlopen(ld->name, RTLD_NOW | RTLD_GLOBAL | RTLD_LOCAL);
	if (ld->dlhandle == NULL) {
		ret = errno;
		ck_printf(CK_PASS, "load failed, dlopen(%s) %s\n", ld->name, dlerror());
	} else {
		func = dlsym(ld->dlhandle, "init");
		if (func == NULL) {
			q = strdup(*module);
			p = strrchr(q, '.');
			if (p == NULL)
				p = q;
			else
				p++;
			asprintf(&r, "%s_init", p);
			func = dlsym(ld->dlhandle, q);
			free(r);
			free(q);
		}
		if (func == NULL) {
			ret = errno;
			ck_printf(CK_PASS, "load failed, no \"init\" function\n");
		} else {
			p = func();
			/* XXX: is there a way to check validity of a pointer ? */
			if (p == NULL || memcmp(p, MAGIC, strlen(MAGIC))) {
				ret = EINVAL;
				ck_printf(CK_PASS, "load failed, module has wrong magic number.\n");
			} else {
				return (0);
			}
		}
	}
	free(ld->name);
	if (ld->dlhandle)
		dlclose(ld->dlhandle);
	TAILQ_REMOVE(&load_head, ld, next);
	free(ld);
	return (ret);
}
