/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: main.c,v 1.51 2007/01/16 21:06:44 phk Exp $
 *
 * This is the main entrance where we set up the event engine, parse the
 * options and set the ball rolling by starting events to read the config
 * file.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <err.h>
#include <errno.h>
#include <isc/eventlib.h>
#include <configkit.h>

#include "ntpns_cfg.h"
#include "ntpns.h"
#include "internal.h"

mod_init_t	oncore_init;
mod_init_t	dcf77_init;
mod_init_t	ntpv4_init;


int dopt, Dopt;

evContext C;

struct ck_tree *cft;

void
fhexdump (FILE *f, u_int len, const void *p)
{
	const u_char *s;
	u_int i;

	s = p;
	for(i = 0; i < len; i++)
		fprintf(f, "%02x", s[i]);
	fprintf(f, "\n");
}

static void
Usage(const char *progname)
{

	errx(1, "Usage: %s [-d] [-f config-file]", progname);
}

int
cfg_isc_debug(CK_ARG, unsigned *arg_0) 
{
	static unsigned debug;

	if (dopt == 0) {
		ck_printf(CK_PASS, "isc_debug only valid in debugging mode");
		return (0);
	}
	ck_UintVar(CK_PASS, arg_0, &debug, 0);
	evSetDebug(C, debug, stderr);
	return (0);
}

int
main(int argc, char **argv)
{
	int i;
	FILE *fd;

	setbuf(stdout, NULL);

	sys_able = ABLE_STARTUP;

	if (getuid() != 0) {
		fprintf(stderr, "Running as non-root, setting '-d' and '-D'\n");
		dopt++;
		Dopt++;
	}

	i = evCreate(&C);
	if (i)
		err(1, "evCreate()");
	i = evSetOption(NULL, "monotime", 1);
	if (i)
		err(1, "evSetOption(NULL, monotime,1)");

	cft = ck_TreeNew();
	ck_TreeAdd(cft, &config_telnet);
	ck_TreeAdd(cft, &config_exit);
	ck_TreeAdd(cft, &config_close);
	ck_TreeAdd(cft, &config_load);
	ck_TreeAdd(cft, &config_record);
	ck_TreeAdd(cft, &config_rrrd);
	ck_TreeAdd(cft, &config_msg);

	init_log();
	init_shmem();
	init_pll();
	init_pllmath();
	init_source();
	init_watch();

	oncore_init();
	dcf77_init();
	ntpv4_init();

	while((i = getopt(argc, argv, "dDf:")) != -1) {
		switch (i) {
		case 'd':
			dopt++;
			break;
		case 'f':
			fd = fopen(optarg, "r");
			if (fd == NULL)
				err(1, optarg);
			ConfigSource(stderr, fd, -1, 0);
			break;
		case 'D':
			Dopt++;
			break;
		default:
			Usage(argv[0]);
		}
	}

	if (dopt)
		ck_TreeAdd(cft, &config_isc_debug);

	if (argc != optind)
		Usage(argv[0]);

	if (dopt) {
		ConfigSource(stderr, stdin, -1, CONFIG_PROMPT);
		if (dopt > 1)
			evSetDebug(C, dopt - 1, stdout);
	} else {
		daemon(1, 1);
	}

	run_shmem();

	i = evMainLoop(C);
	if (i && errno != ENOENT)
		err(1, "evMainLoop()");

	debug(1, "We're done...\n");

	return (0);
}
