/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: pll.c,v 1.33 2005/12/30 10:27:45 phk Exp $
 *
 * PLL handling.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <configkit.h>
#include <err.h>
#include <fcntl.h>
#include <errno.h>
#include <isc/eventlib.h>
#include <sys/queue.h>
#include <sys/timex.h>
#include <math.h>

#include "ntpns.h"
#include "pll.h"
#include "internal.h"

struct logh *L_pll;

int clock_step;

static void
dump_timex(struct timex *tx)
{

	Log(L_pll, 
	 "modes %x status %x offset %ld freq %.3e maxerror %ld esterror %ld constant %ld prec %ld tol %ld",
	    tx->modes, tx->status, tx->offset, tx->freq / (65536.0 * 1e6),
	    tx->maxerror, tx->esterror, tx->constant, tx->precision,
	    tx->tolerance);
}

void
kern_freq(double f)
{
	struct timex tx;
	int i;

	if (Dopt)
		return;

	/* Once we get here first time, startup is complete */
	sys_able &= ~ABLE_STARTUP;

	memset(&tx, 0, sizeof tx);
	tx.modes = MOD_NANO | MOD_FREQUENCY | MOD_OFFSET | MOD_STATUS;
	tx.status = STA_PLL;

	/* Tell the kernel about the leap second status */
	if (sys_leap == SRC_LEAP_INSERT)
		tx.status |= STA_INS;
	else if (sys_leap == SRC_LEAP_DELETE)
		tx.status |= STA_DEL;

	/*
	 * Clip frequency to +/- 500ppm
	 */
	if (f > 500e-6)
		f = 500e-6;
	else if (f < -500e-6)
		f = -500e-6;

	/* Frequency unit is parts of 2^16 of PPM */
	tx.freq = f * (65536.0 * 1e6);

	record("KERNPLL", 0, &tx, sizeof tx);
	i = ntp_adjtime(&tx);
	if (i != 0)
		LogErr(Lwarn, "ntp_adjtime()");
	dump_timex(&tx);
}

void
pllstep(double dd)
{
	struct timespec ts, tsu;
	double d;
	int i;

	if (Dopt)
		return;
	record("PLLSTEP", 0, &dd, sizeof dd);
	d = -dd;
	tsu = evUTCTime();
	if (d < 0) {
		d = -d;
		ts.tv_sec = d;
		d -= ts.tv_sec;
		ts.tv_nsec = d * 1e9;
		ts = evSubTime(tsu, ts);
	} else {
		ts.tv_sec = d;
		d -= ts.tv_sec;
		ts.tv_nsec = d * 1e9;
		ts = evAddTime(tsu, ts);
	}
	i = clock_settime(CLOCK_REALTIME, &ts);
	Log(Linfo, "PLL step (%d) %.09f %d.%09ld -> %d.%09ld",
	    i, dd, tsu.tv_sec, tsu.tv_nsec,
	    ts.tv_sec, ts.tv_nsec);
	if (i != 0)
		LogErr(Lwarn, "clock_settime()");
	clock_step++;
}

void
pll(double y)
{

	kern_freq(pll_math(y));
}

void
pllnone()
{

	kern_freq(pll_poll());
}

const char *
init_pll(void)
{

	L_pll = LogNewHandle("pll");
	return (MAGIC);
}
