/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: shmem.c,v 1.12 2007/05/08 08:57:06 phk Exp $
 *
 * Shared memory interface for publishing info to other processes (like snmpd)
 * XXX: snmpd should support AgentX or similar so this wouldn't be necessary.
 *
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <isc/eventlib.h>
#include "ntpns.h"
#include "internal.h"
#include "shmem.h"

static const char *shmem_name = SHMEM_FILENAME;
static int shmem_fd = -1;
static u_int shmem_len = 0;
static void *shmem_ptr = NULL;

static TAILQ_HEAD(, shmem) shmem_head = TAILQ_HEAD_INITIALIZER(shmem_head);

void *
shmem_alloc(u_int length, u_int tag)
{
	struct shmem *sh;
	u_int i, j, nshl;

	/* Reserve 10 MB of address space for shared memory */
	if (shmem_ptr == NULL) {
		shmem_ptr = mmap(NULL, 10*1024*1024,
		    PROT_NONE, MAP_ANON, -1, 0);
		assert(shmem_ptr != MAP_FAILED);
	}

	sh = malloc(sizeof *sh);
	if (sh == NULL) {
		LogErr(Lwarn,
		     "Could not open shmem file %s", shmem_name);
		return (NULL);
	}
	sh->length = length + sizeof *sh;
	sh->tag = tag;
	TAILQ_INSERT_TAIL(&shmem_head, sh, next);

	if (shmem_fd < 0) {
		unlink(shmem_name);
		shmem_fd = open(shmem_name, O_RDWR | O_CREAT | O_TRUNC, 0644);
		if (shmem_fd < 0) {
			LogErr(Lwarn,
			     "Could not open shmem file %s", shmem_name);
			return (NULL);
		}
	}

	/* New length */
	nshl = shmem_len + sh->length;

	/* Round up to page border */
	i = getpagesize();
	j = (nshl + i - 1) & ~(i - 1);

	ftruncate(shmem_fd, j);

	/* Map it */
	shmem_ptr = mmap(shmem_ptr, j, PROT_READ|PROT_WRITE,
	    MAP_SHARED | (shmem_ptr == NULL ? 0 : MAP_FIXED),
	    shmem_fd, 0);
	if (shmem_ptr == MAP_FAILED) {
		LogErr(Lwarn, "mmap(2) shmem failed");
		return (NULL);
	}

	/* Create new item */
	sh->ptr = (void*)((u_char *)shmem_ptr + shmem_len);
	memcpy(sh->ptr, sh, sizeof *sh);

	/* Update shmem_len */
	Log(Linfo, "SHMEM tag %x len %d at %d/%x",
	    tag, length, shmem_len, shmem_len);
	shmem_len += sh->length;
	return (sh->ptr + 1);
}

void
shmem_free(void *ptr __unused)
{
	/* XXX: TBD */
}

static struct ntpns_id *id;

const char *
init_shmem(void)
{

	id = shmem_alloc(sizeof *id, ID_TAG);
	id->pid = getpid();
	strcpy(id->magic, MAGIC);
	id->major = NTPNS_MAJOR;
	id->minor = NTPNS_MINOR;
	return (MAGIC);
}

void
run_shmem(void)
{

	id->pid = getpid();
}
