/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: telnet.c,v 1.24 2005/09/09 20:13:09 phk Exp $
 *
 * Telnet protocol engine for OaM access.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <err.h>
#include <fcntl.h>
#include <errno.h>
#include <isc/eventlib.h>
#include <sys/queue.h>
#include <sys/socket.h>
#include <sys/endian.h>
#include <math.h>
#include <netdb.h>
#include <configkit.h>

#include "ntpns.h"
#include "ntpns_cfg.h"
#include "internal.h"

static int telnetenable = 0;
static int telnetsock = -1;
static evConnID cid;
static char *telnethost;
static char *telnetport;

static void
telnet_connect(evContext ctx __unused, void *uap __unused, int fd, const void *la __unused, int lalen __unused, const void *ra __unused, int ralen __unused)
{
	
	/* XXX: log telnet connections src+dst IP */
	ConfigSource(NULL, NULL, fd, CONFIG_PROMPT);
}

static void
resolved(CK_ARG, void *priv __unused, int error, struct addrinfo *ai)
{
	int s, i;

	if (error) {
		ck_printf(CK_PASS,
		    "failed to resolve address for record (%d)\n", error);
		return;
	}

	s = GetBoundSocket(NULL,
	    ai->ai_family, ai->ai_socktype, ai->ai_protocol,
	    ai->ai_addr, ai->ai_addrlen);
	if (s == -1)
		return;
	i = 1;
	i = setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &i, sizeof i);
	if (i < 0) 
		warn("setsockopt(SO_REUSEADDR) failed");
	i = evListen(C, s, 10, telnet_connect, NULL, &cid);
	if (i < 0) {
		close (s);
		err(1, "evListen failed");
	} else {
		telnetsock = s;
	}
}

static void
telnet_reconf(CK_ARG)
{
	int i;

	if (telnetsock >= 0) {
		evCancelConn(C, cid);
		close(telnetsock);
		telnetsock = -1;
	}
	if (telnetenable) {
		i = resolver(CK_PASS,
		    telnethost,
		    telnetport == NULL ? "ntp" : telnetport,
		    PF_INET, SOCK_STREAM,
		    AI_CANONNAME | AI_PASSIVE, resolved, NULL);
		if (i < 0)
			ck_printf(CK_PASS, "DNS resolver failed\n");
	}
}

int
cfg_telnet_enable(CK_ARG)
{

	if (cfa->flag & CK_NO) {
		telnetenable = 0;
		telnet_reconf(cfa);
		return (0);
	}
	if (cfa->flag & CK_CONF) {
		if (telnetenable)
			ck_printf(CK_PASS, "telnet enable\n");
		return (0);
	}
	if (cfa->flag & CK_SHOW) {
		ck_printf(CK_PASS, "don't know how to yet\n");
		return (0);
	}
	telnetenable = 1;
	telnet_reconf(cfa);
	return (0);
}

int
cfg_telnet_host(CK_ARG, const char **host)
{

	if (ck_WordVar(CK_PASS, host, &telnethost))
		telnet_reconf(cfa);
	return (0);
}

int
cfg_telnet_port(CK_ARG, const char **port)
{

	if (ck_WordVar(CK_PASS, port, &telnetport))
		telnet_reconf(cfa);
	return (0);
}
