/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: ntpv4.h,v 1.33 2007/01/18 09:49:43 phk Exp $
 */

#define WKP		123

/*
 * Our internal representation of a NTP packet
 */
struct ntp_packet {
	uint8_t		leap;
	uint8_t		version;
	uint8_t		mode;
	uint8_t		stratum;
	uint8_t		poll;
	int8_t		prec;
	double		delay;
	double		disp;
	uint8_t		refid[4];
	struct ts	ts[5];
	int		ours;
};

struct instance {
	u_int			magic;
#define NTPV4_MAGIC		0x1231231
	struct ck_instance	cki;
	struct ck_instancehead	servers;
	u_long			unit;

	/* packet.c */
	char			*listen;
	int			socket;
	evFileID		socket_fid;
	int			userland_ts;

	uint			filter;
#define NTPV4_FILTER_FRAC0	(1 << 0)

};

struct partner;

struct configured {
	u_int			magic;
#define CONF_MAGIC		0x2434342
	struct ck_instance	cki;
	struct instance		*inp;
	LIST_ENTRY(configured)	next;
	char			*name;
	evTimerID		tid;
	LIST_HEAD(,partner)	partner;
	uint			poll;
	double			votes;
};

#define TAG_PARTNER 0x50415254

struct partner {	/* ... in time */
	u_int			magic;
#define PARTNER_MAGIC	0x8383832
	u_int			generation;
	struct instance		*inp;
	/* partner.c fields */
	LIST_ENTRY(partner)	hash;
	TAILQ_ENTRY(partner)	lru;
	uint32_t		ip;		/* should be opaque */
	int			iswkp;
	int			err;
	time_t			last;
	struct sockaddr_in	addr;
	evTimerID		tid;
	uint64_t		originate;

	struct source		*sid;

	void			*cookie;
	u_int			cookielen;

	int			flag;
#define P_F_UPDATE		1
#define P_F_NEVER		2

	LIST_ENTRY(partner)	config;
	struct configured	*configured;

	struct ntp_packet	packet;

	double			delta;
};

extern struct logh *lntpv4, *lsocket, *lpacket, *lpartner;

/* packet.c */
int OpenNTPSocket(evContext ctx, struct instance *ip);
struct msghdr * socket_prepare(struct partner *p, void *buf, u_int len);

/* proto.c */
void RcvPacket(struct instance *ip, void *packet, int len, struct sockaddr_in *s, struct ts *ts, void *cookie, u_int cookielen, u_char *ref);
void poll_timer(evContext ctx, void *uap, struct timespec due, struct timespec inter);
#ifdef CK_ARG
void PrintPacket(CK_ARG, struct partner *pr, int how);
void PrintPacketHeader(CK_ARG);
#endif

/* partner.c */
struct partner *LookupPartner(time_t sec, uint32_t ip, int wkport);
void DeletePartner(struct partner *);
struct configured *NewServer(struct instance *ip, const char *name);
void DeleteConfigured(struct configured *cp);
