/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: phash.c,v 1.3 2005/09/10 17:39:06 phk Exp $
 *
 * Calculate as good a hash function for Oncore messages as we can find.
 * Each (better) solution will be written to ../oncore_hash.c as it is
 * found.
 *
 * The program can be terminated at any time.
 */


#include <stdio.h>
#include <stdlib.h>
#include <err.h>

static int ncode;
static char code[100][2];
static unsigned res[100];

static int
cmp_u(const void *a, const void *b)
{
	const unsigned *au, *bu;

	au = a;
	bu = b;
	return (*au - *bu);
}

static void
Try(unsigned a, unsigned b)
{
	int i, j, n;
	unsigned x, l, c, d;
	static unsigned y, bx = 1000;
	FILE *fo;

	for (d = 0; d < 28; d++) {
	for (c = 0; c < 28; c++) {
		for (i = 0; i < ncode; i++) {
			res[i] = ((code[i][0] * a) >> d) ^
			    ((code[i][1] * b) >> c);
		}
		for (j = 7; j >= 0; j--) {
			y++;
			for (i = 0; i < ncode; i++)
				res[i] &= ((1 << j) - 1);
			x = res[ncode - 1] - res[0];
			if (x >= bx)
				continue;
			qsort(res, ncode, sizeof res[0], cmp_u);
			n = 0;
			for (i = 0; i < ncode - 1; i++)
				if (res[i] == res[i + 1])
					n++;
			if (n != 0) 
				break;
			x = res[ncode - 1] - res[0];
			if (res[ncode - 1] > res[0] && x >= bx)
				continue;
			l = res[0];
			fo = fopen("../oncore_hash.h", "w");
			fprintf(fo, "/* This file is in the public domain */\n");
			fprintf(fo, "/* This file is generated by oncore/Phash/phash.c */\n");
			fprintf(fo, "\n");
			fprintf(fo, "#define N_MASH %d\n", x + 2);
			fprintf(fo, "\n");
			fprintf(fo, "static __inline unsigned\n");
			fprintf(fo, "msg_mash(const void *s)\n");
			fprintf(fo, "{\n");
			fprintf(fo, "\tconst unsigned char *p = s;\n");
			fprintf(fo, "\tunsigned i;\n");
			fprintf(fo, "\n");
			fprintf(fo, "\ti = (\n");
			fprintf(fo, "\t     (\n");
			fprintf(fo, "\t      ((p[2] * 0x%xU) >> %d)\n", a, d);
			fprintf(fo, "\t      ^\n");
			fprintf(fo, "\t      ((p[3] * 0x%xU) >> %d)\n", b, c);
			fprintf(fo, "\t     ) & %d\n", (1 << j) -1);
			fprintf(fo, "\t    ) - %u;\n", l);
			fprintf(fo, "\tif (i >= N_MASH)\n");
			fprintf(fo, "\t\ti = N_MASH - 1;\n");
			fprintf(fo, "\treturn (i);\n");
			fprintf(fo, "}\n");
			fprintf(fo, "\n");
			fprintf(fo, "/*\n");
			for (i = 0; i < ncode; i++) {
				res[i] = ((code[i][0] * a) >> d) ^ ((code[i][1] * b)) >> c;
				res[i] &= ((1 << j) - 1);
				res[i] -= l;
				fprintf(fo, " * %c%c -> %d\n",
				    code[i][0], code[i][1], res[i]);
			}
			fprintf(fo, " */\n");
			fclose(fo);
			printf("%u iterations size %u factor %f (%d %d %d)\n",
			    y, x, x / (double) ncode, c, d, j);
			bx = x;
		}
	}
	}
}

int
main(int argc __unused, char **argv __unused)
{
	int i;
	unsigned u, v;
	char buf[BUFSIZ];

	srandomdev();
	while (fgets(buf, sizeof buf, stdin)) {
		for (i = 0; i < ncode; i++)
			if (code[i][0] == buf[2] &&
			    code[i][1] == buf[3])
				break;
		if (i == ncode) {	
			code[ncode][0] = buf[2];
			code[ncode][1] = buf[3];
			ncode++;
		}
	}
	printf("ncode = %d\n", ncode);
	u = random();
	while (1) {
		v = random();
		Try(u, v);
		u = v;
	}
	return (0);
}
