/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: oncore.c,v 1.64 2005/09/25 22:06:02 phk Exp $
 */

#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <pthread.h>
#include <isc/eventlib.h>

#include <configkit.h>
#include <ntpns.h>

#include "oncore.h"
#include "oncore_cfg.h"

struct logh *loncore_serial, *loncore_msg;

int
oncore_version(CK_ARG)
{
	struct oncore *ip;

	ip = cfa->priv;
	Assert(ip->magic == ONCORE_MAGIC, "Wrong Oncore magic");
	ck_printf(CK_PASS, "%s\n", ip->Cj);
	return (0);
}

int
oncore_show(CK_ARG)
{
	u_int u;
	struct oncore *ip;
	char tbuf[40];
	time_t now;

	ip = cfa->priv;
	Assert(ip->magic == ONCORE_MAGIC, "Wrong Oncore magic");
	ck_printf(CK_PASS, "serial port = %s", ip->serial);
	if (ip->type == M12)
		ck_printf(CK_PASS, "\tserial number = %s", ip->serno);
	ck_printf(CK_PASS, "\nstate = %d", ip->state);
	ck_printf(CK_PASS, "\tvisible/track/lock = %d/%d/%d",
	    ip->vsat, ip->tsat, ip->lsat);
	ck_printf(CK_PASS, "\tdop = %.1f [m]\n", ip->dop * .1);
	if (ip->type == M12) {
		ck_printf(CK_PASS, "raim_limit %.9f/%.9f mask angle %d/%d\n",
		    ip->cf_raim_limit, ip->raim_limit * 1e-7,
		    ip->cf_mask_angle, ip->mask_angle);
	}
	strftime(tbuf, sizeof tbuf, "%Y-%m-%d %H:%M:%S", &ip->tm);
	ck_printf(CK_PASS, "%s.%09d\n", tbuf, ip->fraction);

	if (*ip->leap_info) {
		ck_printf(CK_PASS, "Leap second info: %s\n", ip->leap_info);
		time(&now);
		ck_printf(CK_PASS, "            %d seconds",
		    ip->leap_when - now);
		if (ip->leap_when - now > 86400)
			ck_printf(CK_PASS, " (%d days)",
			    (ip->leap_when - now) / 86400);
		ck_printf(CK_PASS, " from now\n");
		
	}

	ck_printf(CK_PASS, "lat = %u (%.6f), lon = %u (%.6f), ht %u (%.2f)\n",
	    ip->lat, ip->lat / 3.6e6, ip->lon, ip->lon / 3.6e6,
	    ip->ht, ip->ht / 100.0);
	ck_printf(CK_PASS,
	    "http://maps.google.com/maps?ll=%.6f,%.6f&spn=0.03,0.08&t=k\n",
	    ip->lat / 3.6e6, ip->lon / 3.6e6);
	if (ip->type == M12) {
		ck_printf(CK_PASS, "flat = %u, flon = %u, fht %u\n",
		    ip->flat, ip->flon, ip->fht);
	}
	ck_printf(CK_PASS, "rcv_status = 0x%x =", ip->rcv_status);
	if (ip->type == UT) {
		if (ip->rcv_status & 0x80)
			ck_printf(CK_PASS, " PosPropMode");
		if (ip->rcv_status & 0x40)
			ck_printf(CK_PASS, " PoorGeom");
		if (ip->rcv_status & 0x20)
			ck_printf(CK_PASS, " 3D");
		if (ip->rcv_status & 0x10)
			ck_printf(CK_PASS, " 2D");
		if (ip->rcv_status & 0x08)
			ck_printf(CK_PASS, " AcqSat/PosHold");
		if (ip->rcv_status & 0x04)
			ck_printf(CK_PASS, " DiffFix");
		if (ip->rcv_status & 0x02)
			ck_printf(CK_PASS, " InsufSat");
		if (ip->rcv_status & 0x01)
			ck_printf(CK_PASS, " BadAlmanac");
		ck_printf(CK_PASS, "\n");
	} else if (ip->type == M12) {
		switch((ip->rcv_status & 0xe000) >> 13) {
		case 7: ck_printf(CK_PASS, " 3Dfix"); break;
		case 6: ck_printf(CK_PASS, " 2Dfix"); break;
		case 5: ck_printf(CK_PASS, " Propagate"); break;
		case 4: ck_printf(CK_PASS, " PosHold"); break;
		case 3: ck_printf(CK_PASS, " Acq"); break;
		case 2: ck_printf(CK_PASS, " BadGeom"); break;
		case 1: ck_printf(CK_PASS, " Reserv1"); break;
		case 0: ck_printf(CK_PASS, " Reserv0"); break;
		}
		if (ip->rcv_status & 0x1800)
			ck_printf(CK_PASS, " ???");
		if (ip->rcv_status & 0x0400)
			ck_printf(CK_PASS, " NarrowTrack");
		if (ip->rcv_status & 0x0200)
			ck_printf(CK_PASS, " FastAcq");
		if (ip->rcv_status & 0x0100)
			ck_printf(CK_PASS, " FilterReset");
		if (ip->rcv_status & 0x0080)
			ck_printf(CK_PASS, " ColdStart");
		if (ip->rcv_status & 0x0040)
			ck_printf(CK_PASS, " DiffFix");
		if (ip->rcv_status & 0x0020)
			ck_printf(CK_PASS, " PosLock");
		if (ip->rcv_status & 0x0010)
			ck_printf(CK_PASS, " AutoSurvey");
		if (ip->rcv_status & 0x0008)
			ck_printf(CK_PASS, " InsuffSats");
		switch((ip->rcv_status & 0x0006) >> 1) {
		case 0: ck_printf(CK_PASS, " AntOK"); break;
		case 1: ck_printf(CK_PASS, " AntOverCur"); break;
		case 2: ck_printf(CK_PASS, " AntUnderCur"); break;
		case 3: ck_printf(CK_PASS, " AntNoVolt"); break;
		}
		if (ip->rcv_status & 0x0001)
			ck_printf(CK_PASS, " InternalCode");
		ck_printf(CK_PASS, "\n");

		switch (ip->raim_solution) {
		case 0: ck_printf(CK_PASS, "raim_solution = OK"); break;
		case 1: ck_printf(CK_PASS, "raim_solution = ALARM"); break;
		case 2: ck_printf(CK_PASS, "raim_solution = UNKNOWN"); break;
		default: ck_printf(CK_PASS,
		    "raim_solution = ?? (0x%x)", ip->raim_solution); break;
		}
		switch (ip->raim_status) {
		case 0: ck_printf(CK_PASS,
		    ", raim_status = detection+isolation"); break;
		case 1: ck_printf(CK_PASS,
		    ", raim_status = detection only"); break;
		case 2: ck_printf(CK_PASS,
		    ", raim_status = raim not possible"); break;
		default: ck_printf(CK_PASS,
		    ", raim_status = ?? (0x%x)", ip->raim_status); break;
		}
		/* XXX: convert to SVN */
		ck_printf(CK_PASS,
		    "\nraim_removed = %08x", ip->raim_removed);
		ck_printf(CK_PASS,
		    "\traim_1sigma = %.9f [s]", ip->raim_sigma);
		ck_printf(CK_PASS,
		    "\traim_sawtooth = %d [ns]\n", ip->raim_sawtooth);
		ck_printf(CK_PASS, "clock_bias = %d [ns]", ip->clock_bias);
		ck_printf(CK_PASS, "\tosc_offset = %d [Hz]", ip->osc_offset);
		ck_printf(CK_PASS, "\tosc_temp = %.1f [C]\n",
		    ip->osc_temp * .5);

		ck_printf(CK_PASS, "utc_status = 0x%x =", ip->utc_status);
		if (ip->utc_status & 0x80)
			ck_printf(CK_PASS, " enabled");
		if (ip->utc_status & 0x40)
			ck_printf(CK_PASS, " decoded");
	}
	ck_printf(CK_PASS, "\tutc_offset = %d\n", ip->utc_offset);
	ck_printf(CK_PASS, "site_survey = %d (~%d sec left)\n",
	    ip->site_survey, ip->site_survey_timer);
	ck_printf(CK_PASS, "Sat Dopler Elev Azi Health Mode SigStr IODE Status Offset\n");
	for (u = 0; u < NSAT; u++) {
		struct persat *ps;

		ps = &ip->persat[u];
		if (ps->health == 0 && ps->mode == 0 && ps->signal_strength == 0)
			continue;
		if (ps->health == 4 && ps->signal_strength == 0)
			continue;
		ck_printf(CK_PASS, "%3d %6d %4d %3d   %02x    %2x   %3d    %3d  %4x  %.9f\n", u,
		    ps->doppler, ps->elevation, ps->azimuth, ps->health,
		    ps->mode, ps->signal_strength, ps->iode, ps->status,
		    ps->offset * 1e-9);
#if 0
		if (ps->status & 0xe000)
			ck_printf(CK_PASS, " ???");
		if (ps->status & 0x1000)
			ck_printf(CK_PASS, " NarrowSearch");
		if (ps->status & 0x0800)
			ck_printf(CK_PASS, " UsedForTime");
		if (ps->status & 0x0400)
			ck_printf(CK_PASS, " DiffCorrAvail");
		if (ps->status & 0x0200)
			ck_printf(CK_PASS, " InvalidData");
		if (ps->status & 0x0100)
			ck_printf(CK_PASS, " ParityError");
		if (ps->status & 0x0080)
			ck_printf(CK_PASS, " UsedForPosFix");
		if (ps->status & 0x0040)
			ck_printf(CK_PASS, " SatMomentum");
		if (ps->status & 0x0020)
			ck_printf(CK_PASS, " SatAntiSpoof");
		if (ps->status & 0x0010)
			ck_printf(CK_PASS, " SatUnHealty");
#endif
	}
	return (0);
}

int
cfg_oncore_almanac(CK_ARG, const char **almanac)
{
	struct oncore *ip;

	ip = cfa->priv;
	Assert(ip->magic == ONCORE_MAGIC, "Wrong Oncore magic");
	ck_WordVar(CK_PASS, almanac, &ip->almanac_file);
	return (0);
}

int
cfg_oncore_mask_angle(CK_ARG, uint *mask_angle)
{
	struct oncore *ip;

	ip = cfa->priv;
	Assert(ip->magic == ONCORE_MAGIC, "Wrong Oncore magic");
	ck_UintVar(CK_PASS, mask_angle, &ip->cf_mask_angle, 10);
	return (0);
}

int
cfg_oncore_raim_limit(CK_ARG, double *raim_limit)
{
	struct oncore *ip;

	ip = cfa->priv;
	Assert(ip->magic == ONCORE_MAGIC, "Wrong Oncore magic");
	ck_RealVar(CK_PASS, raim_limit, &ip->cf_raim_limit, 1e-6, "%.9f");
	return (0);
}

int
cfg_oncore_pps_setup(CK_ARG)
{
	struct oncore *ip;

	ip = cfa->priv;
	cfa->priv = ip->pps;
	return (0);
}

int
cfg_oncore_source_setup(CK_ARG)
{
	struct oncore *ip;

	ip = cfa->priv;
	Assert(ip->magic == ONCORE_MAGIC, "Wrong oncore magic");
	cfa->priv = ip->sid;
	return (0);
}

struct ck_instance *
oncore_new(CK_ARG, u_int unit)
{
	struct oncore *ip;

	ip = calloc(sizeof *ip, 1);
	if (ip == NULL)
		return (NULL);
	ip->unit = unit;
	ip->serial_fd = -1;
	ip->cf_raim_limit = 1e-6;
	ip->cf_mask_angle = 10;
	PpsCreate(&ip->pps, 1);
	ip->sid = SourceAdd("oncore_%u", unit);
	SourceRefid(ip->sid, "GPS\0");
	SourceParam(ip->sid, SRC_UPDATE_RATE, 1.0);
	SourceParam(ip->sid, SRC_LOS_TIMEOUT, 15.0);
	evInitID(&ip->reset_timer);
	oncore_state(ip, ST_CONFIG);
	ip->magic = ONCORE_MAGIC;
	ip->cip.obj = ip;
	return (&ip->cip);
}

mod_init_t oncore_init;

const char *
oncore_init(void)
{
	static int once = 0;

	if (once) { 
		warnx("oncore module already loaded");
		return (NULL);
	} else {
		once = 1;
	}

	ck_TreeAdd(cft, &oncore_config);

	loncore_serial = LogNewHandle("oncore_serial");
	loncore_msg = LogNewHandle("oncore_msg");

	if (init_oncore_tables() < 0)
		return (NULL);

	return (MAGIC);
}
