/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: oncore.h,v 1.46 2005/09/12 09:03:17 phk Exp $
 */

#include <sys/queue.h>

struct oncore;
struct oncore_msg;
typedef void msgin_t(struct oncore *ip, struct oncore_msg *mp);

#define NSAT		33		/* 1-32 */
#define NCHAN		12
#define NALMANAC	34
#define LALMANAC	33

struct persat {
	/* from @@Bb */
	int	doppler;
	int	elevation;
	int	azimuth;
	int	health;
	/* from @@Ha */
	int	mode;
	int	signal_strength;
	int	iode;
	int	status;
	/* from @@Hn */
	int	offset;
};

enum oncore_state {
	ST_UNCONF = 0,
	ST_CONFIG,
	ST_START,
	ST_LISTEN,
	ST_CONTACT,
	ST_POLL_STATUS,
	ST_RESET,
	ST_RESET_WAIT,	
	ST_TEST,
	ST_TEST_WAIT,
	ST_SICK,
	ST_INIT,
	ST_SYNC_WAIT,
	ST_SYNC
};

struct oncore {
	u_int			magic;
#define ONCORE_MAGIC		0xe338182
	struct ck_instance cip;
	u_int			unit;

	enum {UNKNOWN, UT, M12}	type;
	enum oncore_state	state;
	evTimerID		reset_timer;

	struct source		*sid;

	char			*serial;
	int			serial_fd;
	evFileID		serial_fid;
	u_char			serial_buf[1024];
	int			serial_idx;

	uint			raim_limit;
	uint			mask_angle;
	double			cf_raim_limit;
	uint			cf_mask_angle;

	struct ppsdev		*pps;

	struct persat		persat[NSAT];

	/* from @@Gj */
	char			leap_info[60];
	time_t			leap_when;

	/* from @@Ha */
	struct tm	tm;
	time_t		t;
	int		fraction;
	int		flat, flon, fht;
	int		lat, lon, ht;
	int		dop, vsat, tsat, lsat;
	u_int		rcv_status;
	int16_t		clock_bias;
	int		osc_offset;
	int		osc_temp;
	int		utc_status;
	int		lutc_status;
	int		gmt_sign;
	int		gmt_hour;
	int		gmt_min;
	char		serno[7];
	struct ts	ts;
	double		delta;

	/* from @@Hn */
	u_int		raim_solution;
	u_int		raim_status;
	uint32_t	raim_removed;
	double		raim_sigma;
	int		raim_sawtooth;

	/* from @@Cb */
	enum {
	    ALMANAC_COLLECT = 0,
	    ALMANAC_LOAD,
	    ALMANAC_LOADED,
	    ALMANAC_SAVE,
	    ALMANAC_DONE
	 }		almanac_work;
	pthread_t	almanac_thread;

	char		*almanac_file;
	u_char		*almanac;
	u_char		*almanac_ptr;
	int		almanac_n;

	/* from @@At: Site survey */
	int		site_survey;
	int		site_survey_timer;

	/* from @@Bo */
	int		utc_offset;

	/* from @@Cj */
	char		Cj[200];
};

extern struct logh *loncore_serial, *loncore_msg;

/* oncore_serial.c */
void oncore_sendmsg(struct oncore *ip, const void *msg, u_int len);
void oncore_initialize(struct oncore *ip);

/* oncore_msg.c */

struct oncore_msg {
	int		rq_len, rs_len;
	char const	*rq_format, *rs_format;
	msgin_t		*handler;
};

int init_oncore_tables(void);

extern char const *oncore_response[];
extern char const *oncore_request[];
extern struct oncore_msg **oncore_msgs;

#define DEFAULT_ALMANAC_FILE "/var/tmp/oncore_almanac.bin"

/* oncore_state.c */
void oncore_state(struct oncore *ip, enum oncore_state state);
void oncore_reset_timer(struct oncore *ip);
