/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: quorum.c,v 1.8 2005/08/23 09:11:25 phk Exp $
 *
 * Quorum based clock selection.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <configkit.h>
#include <err.h>
#include <fcntl.h>
#include <errno.h>
#include <isc/eventlib.h>
#include <sys/queue.h>
#include <math.h>

#include "ntpns.h"
#include "source.h"
#include "pll.h"

struct vote {
	LIST_ENTRY(vote)	list;
	double			offset;
	double			votes;
	struct source 		*sid;
};

struct clockselect {
	struct vote		vote[2];
};

LIST_HEAD(, vote)		votes;

static struct source *
quorum(evContext ctx __unused, struct source *cur __unused, clk_sel_e action)
{
	struct source *sid;
	struct vote *vp, *vp2;
	struct clockselect *cs;
	struct timeline *tl;
	int i;
	double a, ma;

	Abort("Not in use");
	if (action == DEINIT) {
		TAILQ_FOREACH(sid, &sources, list) {
			free(sid->clk_sel);
			sid->clk_sel = NULL;
		}
		return (NULL);
	}
	if (action == INIT)
		return (NULL);
	LIST_INIT(&votes);
	TAILQ_FOREACH(sid, &sources, list) {
		if (sid->flags & SRC_F_DELETE) {
			free(sid->clk_sel);
			sid->clk_sel = NULL;
			continue;
		}
		if (sid->clk_sel == NULL) 
			sid->clk_sel = malloc(sizeof *sid->clk_sel);
		if (sid->clk_sel == NULL) 
			continue;	/* XXX: whine */

		cs = sid->clk_sel;
		tl = &sid->t[sid->hist0];
		if (!tl->v[0])
			continue;
		if (tl->a > .128)
			continue;
		a = tl->a + sid->dispersion;
		cs->vote[0].votes = sid->votes;
		cs->vote[0].sid = sid;
		cs->vote[0].offset = tl->d[0] - a;
		cs->vote[1].votes = -sid->votes;
		cs->vote[1].sid = sid;
		cs->vote[1].offset = tl->d[0] + a;
		for (i = 0; i < 2; i++) {
			if (LIST_EMPTY(&votes)) {
				LIST_INSERT_HEAD(&votes, &cs->vote[i], list);
				continue;
			}
			vp2 = NULL;
			LIST_FOREACH(vp, &votes, list) {
				vp2 = vp;
				if (vp->offset < cs->vote[i].offset)
					continue;
				LIST_INSERT_BEFORE(vp, &cs->vote[i], list);
				break;
			}
			if (vp == NULL)
				LIST_INSERT_AFTER(vp2, &cs->vote[i], list);
		}
	}
	printf("Quorum:\n");
	ma = a = 0;
	vp2 = NULL;
	LIST_FOREACH(vp, &votes, list) {
		a += vp->votes;
		if (a > ma) {
			ma = a;
			vp2 = vp;
		}
		printf("%12.9f %3.0f %3.0f %-20s",
		     vp->offset, a, vp->votes, vp->sid->name);
		for (i = 0; i < a; i++)
			putchar('#');
		printf("\n");
	}
	if (vp2 != NULL) {
		return(vp2->sid);
	} else {
		return(NULL);
	}
}

ref_init_t init;

const char *
init(evContext ctx)
{

	ClockSelect(ctx, quorum);
	return (MAGIC);
}
