/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 *
 * $Id: dcf77_blame.c,v 1.7 2005/10/04 07:41:51 phk Exp $
 *
 */

#include <stdio.h>
#include <string.h>
#include <configkit.h>
#include <sys/types.h>
#include <sys/time.h>
#include <isc/eventlib.h>
#include "ntpns.h"
#include "dcf77.h"


static struct dcf77_template {
	int	midnight;	/* Time pattern minute before midnight */
	int	field;		/* Which integer field this is part of */
	int	weight;		/* The weight in the integer field */
	int	parity;		/* Which parity it to contribute to */
} templ[59] = {
	{ /*  0 */  0,  0, 0, 0 },
	{ /*  1 */  0,  0, 0, 0 },
	{ /*  2 */  0,  0, 0, 0 },
	{ /*  3 */  0,  0, 0, 0 },
	{ /*  4 */  0,  0, 0, 0 },
	{ /*  5 */  0,  0, 0, 0 },
	{ /*  6 */  0,  0, 0, 0 },
	{ /*  7 */  0,  0, 0, 0 },
	{ /*  8 */  0,  0, 0, 0 },
	{ /*  9 */  0,  0, 0, 0 },
	{ /* 10 */  0,  0, 0, 0 },
	{ /* 11 */  0,  0, 0, 0 },
	{ /* 12 */  0,  0, 0, 0 },
	{ /* 13 */  0,  0, 0, 0 },
	{ /* 14 */  0,  0, 0, 0 },
	{ /* 15 */  0,  1, 1, 0 },	/* R */
	{ /* 16 */  0,  2, 1, 0 },	/* A1 (dst) */
	{ /* 17 */  0,  3, 1, 0 },	/* Z1 */
	{ /* 18 */  0,  3, 2, 0 },	/* Z2 */
	{ /* 19 */  0,  4, 1, 0 },	/* A2 (leap) */
	{ /* 20 */  0,  5, 1, 0 },	/* S */
	{ /* 21 */  1,  6, 1, 1 },	/* Minute ones */
	{ /* 22 */  0,  6, 2, 1 },
	{ /* 23 */  0,  6, 4, 1 },
	{ /* 24 */  1,  6, 8, 1 },
	{ /* 25 */  1,  7, 1, 1 },	/* Minute tens */
	{ /* 26 */  0,  7, 2, 1 },
	{ /* 27 */  1,  7, 4, 1 },
	{ /* 28 */  0,  0, 0, 1 },	/* P1 */
	{ /* 29 */  1,  8, 1, 2 },	/* Hour ones */
	{ /* 30 */  1,  8, 2, 2 },
	{ /* 31 */  0,  8, 4, 2 },
	{ /* 32 */  0,  8, 8, 2 },
	{ /* 33 */  0,  9, 1, 2 },	/* Hour tens */
	{ /* 34 */  1,  9, 2, 2 },
	{ /* 35 */  1,  0, 0, 2 },	/* P2 */
	{ /* 36 */  0, 10, 1, 3 },	/* Day ones */
	{ /* 37 */  0, 10, 2, 3 },
	{ /* 38 */  0, 10, 4, 3 },
	{ /* 39 */  0, 10, 8, 3 },
	{ /* 40 */  0, 11, 1, 3 },	/* Day tens */
	{ /* 41 */  0, 11, 2, 3 },
	{ /* 42 */  0, 12, 1, 3 },	/* Week day */
	{ /* 43 */  0, 12, 2, 3 },
	{ /* 44 */  0, 12, 4, 3 },
	{ /* 45 */  0, 13, 1, 3 },	/* Month ones */
	{ /* 46 */  0, 13, 2, 3 },
	{ /* 47 */  0, 13, 4, 3 },
	{ /* 48 */  0, 13, 8, 3 },
	{ /* 49 */  0, 14, 1, 3 },	/* Month tens */
	{ /* 50 */  0, 15, 1, 3 },	/* Year ones */
	{ /* 51 */  0, 15, 2, 3 },
	{ /* 52 */  0, 15, 4, 3 },
	{ /* 53 */  0, 15, 8, 3 },
	{ /* 54 */  0, 16, 1, 3 },	/* Year tens */
	{ /* 55 */  0, 16, 2, 3 },
	{ /* 56 */  0, 16, 4, 3 },
	{ /* 57 */  0, 16, 8, 3 },
	{ /* 58 */  0,  0, 0, 3 },	/* P3 */
};


/*
 * Sanity check a two digit field
 */
static int
check2f(int f1, int f2, int low, int high, int *val)
{
	int i;

	if (val != NULL)
		*val = -1;
	/*
	 * If the tens-digit is missing, we cannot prove anything with
	 */
	if (f2 == -1)
		return (0);
	i = 10 * f2;
	if (f1 != -1) {
		/*
		 * We can only check the lower limit if we have both
		 * tens and units digit.
		 */
		i += f1;
		if (i < low)
			return (1);
	}
	/*
	 * High limit can be checked with just the tens digit: if that
	 * is too high, adding the units in will only make it worse.
	 */
	if (i > high)
		return (1);

	if (val != NULL)
		*val = i;
	return (0);
}


/*
 * Try as chaply as possible to disprove that we have a valid timegram
 * starting at offset.  We do this by exploiting any bit of knowledge
 * and dependency we can in the timegram, without knowing what the time
 * and date actually is.
 *
 * Named after the ruling business paradigm of the new millenium.
 */

int
blameallocation(struct dcf77 *ip, int offset, int limit, char *why)
{
	int *f, i, j, k, dy, mn, yr, reject;
	struct dcf77_template *tp;

	*why = '\0';
#define FAIL(foo)						\
	do {							\
		sprintf foo;					\
		why += strlen(why);				\
		if (++reject >= limit)				\
			return (reject);			\
	} while (0)
	

	reject = 0;

	/* Check the preamble against the configured pattern */
	for (i = 0; i < NPREAMBLE; i++)
		if (ip->preamble[i] * ip->shiftreg[(offset + i) % 60] < 0)
			FAIL((why, " p%d", i));
		
	/* Check the constant S bit */
	j = ip->shiftreg[(offset + 20) % 60];
	if (j == -1)
		FAIL((why, " s"));

	/* Bit 59 must be absent */
	j = ip->shiftreg[(offset + 59) % 60];
	if (j != 0)
		FAIL((why, " e"));

	/* Z1 and Z2 must be different */
	if (ip->shiftreg[(offset + 17) % 60] *
	    ip->shiftreg[(offset + 18) % 60] > 0)
		FAIL((why, " z"));

	/* A1 and A2 can not be on at the same time.  */
	if (ip->shiftreg[(offset + 16) % 60] > 0 &&
	    ip->shiftreg[(offset + 19) % 60] > 0)
		FAIL((why, " a"));

	/* A1 can only happen on a sunday in the end of the month */
	if (ip->shiftreg[(offset + 16) % 60] > 0 &&
	    (ip->shiftreg[(offset + 41) % 60] < 0 ||
	    ip->shiftreg[(offset + 42) % 60] < 0 ||
	    ip->shiftreg[(offset + 43) % 60] < 0 ||
	    ip->shiftreg[(offset + 44) % 60] < 0))
		FAIL((why, " a1"));

	/* A2 can only happen on the first day in the month */
	if (ip->shiftreg[(offset + 19) % 60] > 0 &&
	    (ip->shiftreg[(offset + 45) % 60] < 0 ||
	    ip->shiftreg[(offset + 46) % 60] > 0 ||
	    ip->shiftreg[(offset + 47) % 60] > 0 ||
	    ip->shiftreg[(offset + 48) % 60] > 0 ||
	    ip->shiftreg[(offset + 49) % 60] > 0))
		FAIL((why, " a2"));

	/*
	 * A1 and A2 can only happen in the first three hours of the day
	 * XXX: We could tighten this even further.
	 */
	if ((ip->shiftreg[(offset + 16) % 60] > 0 ||
	    ip->shiftreg[(offset + 19) % 60] > 0) &&
	    (ip->shiftreg[(offset + 31) % 60] > 0 ||
	    ip->shiftreg[(offset + 32) % 60] > 0 ||
	    ip->shiftreg[(offset + 33) % 60] > 0 ||
	    ip->shiftreg[(offset + 34) % 60] > 0))
		FAIL((why, " ah"));

	/* LSB of minutes must be different from previous minute */
	j = ip->shiftprev[(offset + 21) % 60];
	if (j * ip->shiftreg[(offset + 21) % 60] > 0)
		FAIL((why, " 0"));

	/*
	 * If the LSB of minutes was '1' in previous minute
	 * the next higher bit must have changed, if it was
	 * a '0' it must not.
	 */
	if (j *
	    ip->shiftreg[(offset + 22) % 60] *
	    ip->shiftprev[(offset + 22) % 60] > 0)
		FAIL((why, " 1"));

	/*
	 * Try to disprove the hypothesis of a midnight roll-over,
	 * and record from where in the timegram we can expect it to
	 * be the same as last minute.
	 */
	tp = templ + 21;
	for (j = 21; j < 35; j++, tp++) {
		if (ip->shiftprev[(offset + j) % 60] * tp->midnight >= 0)
			continue;

		/* Last minute was not midnight */

		if (j < 25) {
			/*
			 * Units of minute disproved a rollover, so we know
			 * that tens of minute, hours and day must be the same.
			 * But P1 may be different, so we do a local check
			 * here for the ten of minutes sans P1.
			 */
			for (i = 25; i <= 27; i++)
				if (ip->shiftprev[(offset + i) % 60] *
				    ip->shiftreg[(offset + i) % 60] < 0)
					FAIL((why, " d(%d)", i));
			/* And tell it to go from the hours field */
			i = 29;
		} else if (j < 29) {
			/*
			 * Tens of minute disproved rollover, check hours
			 * and day.
			 */
			i = 29;
		} else if (j < 30) {
			/*
			 * Lowest bit of units of hour disproved midnight,
			 * test rest of hour and day.
			 */
			i = 30;
		} else {
			/*
			 * Hours disproved rollover, check day.  We cannot
			 * check separately on units of hour because the
			 * roll-over value for units of hour varies (9/9/3).
			 */
			i = 36;
		}

		/* Check the bits which should not differ from last minute */
		for (; i <= 58; i++)
			if (ip->shiftprev[(offset + i) % 60] *
			    ip->shiftreg[(offset + i) % 60] < 0)
				FAIL((why, " d(%d)", i));

		break;
	}

	/*
	 * Next we start the numerical checks on the individual fields of
	 * the timegram.  First decode the fields.
	 */

	/* Initialize all fields to zero */
	for (i = 0; i < 17; i++)
		ip->field[i] = 0;

	/* ... and all parities to "OK" */
	for (i = 0; i < 4; i++)
		ip->parity[i] = 1;

	/* Run through template and dissect timegram */
	tp = templ;
	for (i = 0; i < 59; i++, tp++) {
		j = ip->shiftreg[(i + offset) % 60];

		/*
		 * Calculating the even parity, we want to track the
		 * number of '1' bits, but since the multiplication
		 * 'tracks' the negative numbers, we have to flip the
		 * sign.
		 */
		ip->parity[tp->parity] *= -j;

		f = &ip->field[tp->field];
		if (j == 0) {
			*f = -1;
		} else if (j == 1 && *f != -1) {
			*f += tp->weight;

			/* No BCD digit can be higher than 9 */
			if (*f > 9)
				FAIL((why, " b%d", tp->field));
		}
	}

	/*
	 * Invalidate the parity for the field we are currently inserting
	 * into, unless we insert into the last byte of the field.
	 */
	k = templ[(60 + ip->lasti - offset) % 60].parity;
	j = templ[(61 + ip->lasti - offset) % 60].parity;
	if (k == j)
		ip->parity[k] = 0;

	/*
	 * Check the parity fields
	 */
	for (j = 1; j <= 3; j++)
		if (ip->parity[j] == -1)
			FAIL((why, " P%d", j));

	/* Check that minute is 0...59 */
	if(check2f(ip->field[6], ip->field[7], 0, 59, NULL))
		FAIL((why, " m(%d,%d)", ip->field[7], ip->field[6]));

	/* Check that hour is 0...23 */
	if(check2f(ip->field[8], ip->field[9], 0, 23, NULL))
		FAIL((why, " h(%d,%d)", ip->field[9], ip->field[8]));

	/* Check that day is 1...31 */
	if(check2f(ip->field[10], ip->field[11], 1, 31, &dy))
		FAIL((why, " D(%d,%d)", ip->field[11], ip->field[10]));

	/* Check that weekday is 1...7 */
	if (ip->field[12] == 0 || ip->field[12] > 7)
		FAIL((why, " W(%d)", ip->field[12]));

	/* Check that month is 1...12 */
	if(check2f(ip->field[13], ip->field[14], 1, 12, &mn))
		FAIL((why, " M(%d,%d)", ip->field[14], ip->field[13]));

	/*
	 * We need not check that the year is 00...99 because the "x <= 9"
	 * check in the loop above has already done that for us, we do 
	 * however want to know the value of the year field for the
	 * next check, so do it anyway for that side effect.
	 */
	if(check2f(ip->field[15], ip->field[16], 0, 99, &yr))
	       FAIL((why, " y(%d,%d)", ip->field[16], ip->field[15]));

	/* Check length of month isn't exceeded */
	switch (mn) {
	case 2:
		/*
		 * Leap-years are SO MUCH FUN!  We do not have access to the
		 * century in the DCF77 signal, so we cannot know if the
		 * century is divisible by four.  Be generous and accept
		 * feb29 in all years divisible by four
		 */
		if (yr != -1 && (yr % 4) != 0)
			i = 28;
		else
			i = 29;
		if (dy > i)
			FAIL((why, " c(%d,%d)", dy, mn));
		break;
	case 4: case 6: case 9: case 11:
		if (dy > 30)
			FAIL((why, " c(%d,%d)", dy, mn));
		break;
	default:
		break;
	}
	return (reject);
}
